#!/usr/bin/env python3

"""Chained CI Pipeline Manager."""

import json
import time
import urllib.parse
from chainedci.config import getConfig
from chainedci.http_adapter import Http
from chainedci.log import log
from chainedci.tools import (raise_ex, get_env_or_raise, token_selector,
                             waiting_cursor)


def _prepare_variables(parameters):
    """Prepare variable list to be sent to gitlab trigger API."""
    return {f'variables[{var}]': val for (var, val) in parameters.items()}


class Pipeline():
    """Pipeline manager."""

    def __init__(self, base_url, token, parameters, ref='master'):
        """Prepare the pipeline."""
        self.base_url = base_url
        self.parameters = _prepare_variables(parameters)
        self.parameters['token'] = token
        self.parameters['ref'] = ref
        self.pipeline_id = None
        self.status = {'status': None,
                       'start': None,
                       'timeout': None,
                       'custom_timeout': None,
                       'previous_time': None,
                       'step_running_timeout': None,
                       'last_response': None}
        self.web_url = None
        self.stop = False

    def loop(self):
        """Start the pipeline and wait for the end."""
        self.start()
        self.wait_for_remote_pipeline()

    def set_custom_timeout(self, timeout):
        """
        Set pipeline specific running timeout.

        :param timeout: The time to wait in running status before
                        stopping with timeout alert
        :type timeout: int
        """
        ini = getConfig('ini')
        if ini['timers']['timeout_retrocompatibility']:
            self.status['step_running_timeout'] = (
                timeout * ini['timers']['retries_sleep'])
        else:
            self.status['step_running_timeout'] = timeout

    def start(self):
        """
        Start the remote pipeline.

        :return: Run result
        :rtype: string
        """
        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        # Send the POST
        url = self.base_url+"/trigger/pipeline"
        log.debug("Pipeline: start - POST to '%s'", url)
        log.debug("Pipeline: start - POST with params '%s'",
                  dict(self.parameters))
        with Http().session.post(url,
                                 data=urllib.parse.urlencode(self.parameters),
                                 headers=headers) as response:

            # Handle success
            if response.status_code == 201:
                last_response = json.loads(response.content)
                self.pipeline_id = last_response['id']
                self._set_status('pending')
                log.debug("Pipeline: start - Pipeline successfully created "
                          "at '%s' (Gitlab server time) "
                          "with id [%s] and status '%s'",
                          last_response['created_at'],
                          last_response['id'],
                          last_response['status'])
                log.info("Pipeline: start - Pipeline created: '%s'",
                         last_response['web_url'])
                self.status['last_response'] = last_response

            # Handle Failure
            elif response.status_code == 404:
                raise_ex(ValueError,
                         (f"Pipeline: run - Failed to start "
                          f"'{url}' with code [404]. "
                          "This may be an issue with 'api' or 'trigger_token' "
                          "values"))
            else:
                raise_ex(ValueError,
                         (f"Pipeline: run - Failed to start "
                          f"'{url}' with code "
                          f"[{response.status_code}] and message "
                          f"'{response.content}'"))

    def get_remote_pipeline_info(self):
        """Get the remote pipeline info."""
        url = f"{self.base_url}/pipelines/{self.pipeline_id}"
        headers = token_selector(url)
        log.debug("Pipeline: get status from '%s'", url)
        with Http().session.get(url, headers=headers) as response:
            # Manage non 200 answers
            if response.status_code not in [200, 401]:
                raise_ex(ValueError,
                         (f"Pipeline: run - Failed to get status at "
                          f"'{url}' with code "
                          f"[{response.status_code}] and message "
                          f"'{response.content}'"))
            # Manage some exceptions where there is no content, it can happen
            if ((response.content == b'' or response.status_code == 401)
                    and self.status['status'] != 'unknown'):
                log.debug("Pipeline: status changed from 'pending' "
                          "to 'unknown'")
                return 'unknown'
            return json.loads(response.content)

    def get_remote_pipeline_status(self):
        """
        Get the remote pipeline status and handle response.

        Possible return of Gitlab CI are: created, waiting_for_resource,
        preparing, pending, running, success, failed, canceled, skipped,
        manual, scheduled.

        For now, created, waiting_for_resource, preparing, skipped, manual,
        and scheduled, are not managed.
        """
        resp = self.get_remote_pipeline_info()
        if resp == 'unknown':
            self._set_status('unknown')
        else:
            log.debug("Pipeline: last status = '%s'", resp['status'])
            if resp['status'] == 'running':
                # Rollback on timer if we are in unknown state
                if self.status['status'] == 'unknown':
                    log.debug("Pipeline: status changed from 'unknown' "
                              "to 'running'")
                    self._set_status('running',
                                     self.status['previous_time'])
                # Reset timers if we move from pending to running
                elif self.status['status'] == 'pending':
                    log.debug("Pipeline: status changed from 'pending' "
                              "to 'running'")
                    self._set_status('running')
            elif resp['status'] in ['failed', 'success', 'canceled']:
                ini = getConfig('ini')
                if ini['log']['level'] == 'INFO':
                    print('\n')
                log.info("Pipeline: stopping, status changed to '%s'",
                         resp['status'])
                self.stop = True
            self.status['last_response'] = resp
            self.status['status'] = resp['status']

    def _set_status(self, status, start_time=None):
        """
        Update status state and time.

        :param status: The status
        :type name: string
        :param start_time: The epoch to force for the start time
        :type name: float
        """
        ini = getConfig('ini')
        if status == 'unknown':
            self.status['previous_time'] = self.status['start']
        start_time = start_time or time.time()
        status_timer = (self.status['step_running_timeout']
                        if (status == 'running' and
                            self.status['step_running_timeout']) else
                        ini['timers'][status])
        self.status['start'] = start_time
        self.status['timeout'] = start_time + status_timer
        self.status['status'] = status

    def wait_for_remote_pipeline(self):
        """Loop to wait pipeline end."""
        log.debug("Pipeline - Loop start")
        ini = getConfig('ini')
        while not self.stop:
            if time.time() < self.status['timeout']:
                log.debug("Pipeline - Wait %s seconds",
                          ini['timers']['retries_sleep'])
                waiting_cursor(ini['timers']['retries_sleep'])
                self.get_remote_pipeline_status()
            else:
                raise_ex(
                    TimeoutError,
                    (f"Pipeline: Timeout, pipeline {self.web_url} "
                     f"seems stucked in status '{self.status['status']}' "
                     f"from {self.status['timeout'] - self.status['start']} "
                     "seconds")
                )
        log.info("Pipeline - Loop ended with status '%s'",
                 self.status['status'])

    def artifact_url(self, job_name):
        """Get the remote artifact url."""
        ini = getConfig('ini')
        # Warning if not a successful pipeline
        if self.status['status'] != 'success':
            raise_ex(ValueError,
                     "Pipeline - no artifact will be available with "
                     f"that '{self.status['status']}' pipeline")
        headers = {"PRIVATE-TOKEN":
                   get_env_or_raise(ini['auth']['private_token_env'])}
        url = (f"{self.base_url}/pipelines/{self.pipeline_id}/jobs/"
               "?scope[]=success")
        log.debug("Pipeline: get successful jobs from '%s'", url)
        with Http().session.get(url, headers=headers) as response:
            # Manage non 200 answers
            if response.status_code not in [200]:
                raise_ex(ValueError,
                         (f"Pipeline: artifact_url - Failed to get jobs at "
                          f"'{url}' with code "
                          f"[{response.status_code}] and message "
                          f"'{response.content}'"))
            resp = json.loads(response.content)
            log.debug("Pipeline - Job list = %s",
                      [{key: job[key] for key in ['id', 'name', 'status']}
                       for job in resp])
            jobs = sorted(filter(lambda d: d['name'] == job_name, resp),
                          key=lambda i: i['id'])
            if len(jobs) < 1:
                raise_ex(ValueError,
                         (f"Pipeline: artifact_url - Failed to get the jobs "
                          f"list with name '{job_name}'. Get {len(jobs)}, but "
                          "shall have more than '1'."))
            if 'artifacts_file' not in jobs[-1].keys():
                raise_ex(ValueError,
                         (f"Pipeline: artifact_url - No artifact file linked "
                          f"with this job ({jobs[-1]})"))
            url = f"{self.base_url}/jobs/{jobs[-1]['id']}/artifacts"
            log.info("Pipeline - artifact is here '%s'", url)
            return url



#!/usr/bin/env python3

"""Simple Jinja-like parser."""

import re
import logging
from chainedci.config import getConfig


class JinjaParser():
    """Ultra light ansible jinja parser."""

    def __init__(self):
        """Prepare the regex."""
        # pylint: disable=anomalous-backslash-in-string
        self.re_tag = re.compile(".*\\{\\{(.*?)\\}\\}.*")
        # pylint: disable=anomalous-backslash-in-string
        self.re_lookup = re.compile(
            "\\s*lookup\\(\\s*'env'\\s*,\\s*'(.*?)'\\s*\\)")
        # pylint: disable=anomalous-backslash-in-string
        self.re_default = re.compile("\\s*default\\(\\s*'(.*?)'\\s*")

    def parse(self, data, hide_value=False):
        """
        Parse the string.

        :param data: The String to parse
        :type data: string
        :param hide_value: Hide the replaced value
        :type hide_value: boolean
        :return: the string with replaced part
        :rtype: string
        """
        if isinstance(data, str):
            while self.re_tag.match(data):
                log = logging.getLogger('chainedci')
                log.debug("JinjaParser - Complex variable found in '%s'", data)
                matchs = self.re_tag.findall(data)
                for expr in matchs:
                    data = data.replace("{{" + expr + "}}",
                                        str(self.parse_expr(expr, log, hide_value)))
        return data

    def parse_expr(self, expr, log, hide_value):
        """
        Parse one expression.

        :param expr: The epxression to parse
        :type expr: string
        :param log: The logging of object
        :type log: logging
        :param hide_value: Hide the replaced value
        :type hide_value: boolean
        :return: replacement of the expression
        :rtype: string
        """
        ret = ''
        log.debug("JinjaParser - Try to parse '%s'", expr)
        expr_parts = expr.split('|')
        for part in expr_parts:
            if self.re_lookup.match(part):
                ret = self.lookup_env(part, hide_value)
            elif self.re_default.match(part):
                ret = self.default(ret, part, hide_value)
            else:
                raise ValueError(f'Can not parse {part} in {expr}')
        return ret

    def lookup_env(self, part, hide_value=False):
        """
        Lookup for variable in env variables.

        :param part: the lookup('env', 'var') part
        :type part: string
        :param hide_value: Hide the replaced value
        :type hide_value: boolean
        :return: the replaced string
        :rtype: string
        """
        ini = getConfig('ini')
        log = logging.getLogger('chainedci')
        log.debug("JinjaParser - Match lookup in '%s'", part)
        matchs = self.re_lookup.findall(part)
        if matchs[0] in ini['env']:
            log.debug("JinjaParser - Match lookup replace with '%s'",
                      '*****' if hide_value else ini['env'][matchs[0]])
            return ini['env'][matchs[0]]
        log.debug("JinjaParser - No value found, set to ''")
        return ''

    def default(self, prev, part, hide_value=False):
        """
        Set default value.

        :param prev: the default('value', ) part
        :type prev: the previous value
        :param part: the lookup('env', 'var') part
        :type part: string
        :param hide_value: Hide the replaced value
        :type hide_value: boolean
        :return: the replaced string
        :rtype: string
        """
        log = logging.getLogger('chainedci')
        log.debug("JinjaParser - Match default in '%s'", part)
        matchs = self.re_default.findall(part)
        if prev == '' or prev is None:
            log.debug("JinjaParser - set value to default '%s'",
                      '*****' if hide_value else matchs[0])
            return matchs[0]
        return prev



#!/usr/bin/env python3

"""Setup and install chainedci."""

from glob import glob
from os.path import basename
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def readme():
    """Set Readme from file."""
    with open('README.md', encoding="utf-8") as f:
        return f.read()


setup(name='chainedci',
      version='1.5.13',
      description='Chaine Gitlab CI pipelines',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://gitlab.com/Orange-OpenSource/lfn/ci_cd/chained-ci',
      author='Orange OpenSource',
      license='Apache 2.0',
      packages=find_packages(),
      py_modules=[splitext(basename(path))[0] for path in glob('*.py')],
      include_package_data=True,
      scripts=["chainedci/chainedci"],
      install_requires=[
          "ansible-core==2.14.3",
          "GitPython==3.1.31",
          "Jinja2==3.1.2",
          "requests==2.28.2",
          "schema==0.7.5",
          "urllib3 == 1.26.14"
      ],
      setup_requires=["pytest-runner"],
      tests_require=[
          "pytest",
          "pytest-cov",
          "pytest-mock",
          "mock",
          "requests_mock"
      ],
      zip_safe=False)



#!/usr/bin/env python3
"""Chained CI engine."""


from chainedci.config import init_ini, getConfig
from chainedci.log import init_log, log
from chainedci.version import __version__

init_ini()
init_log(getConfig('ini'))
log.info("---- ChainedCI v%s initialization -----", __version__)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0

"""HTTP adapter for chained py.

some simple functions used in the different classes
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar, NoReturn

from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry  import Retry

from chainedci.log import log

def generate_default_max_replies():
    """Generate the default max replies."""
    return Http.default_max_retries


def generate_default_backoff():
    """Generate the defaukt backoff."""
    return Http.default_backoff


@dataclass
class Http():
    """
    Http Class.

    This class is intended to configure requests with timeouts and retries.

    :param max_retries: the max retries number. Will be set to
        Http.default_max_retries if not given.
    :param backoff: the backoff factor. Will be set to Http.default_timeout if
        not given.
    :param session: the session with a right retry strategy configured

    :type max_retries: int
    :type backoff: int
    :type session: :class:`Session`

    The backoff factors allows you to change how long the processes will sleep
    between failed requests. The algorithm is as follows:

    {backoff factor} * (2 ** ({number of total retries} - 1))

    For example, if the backoff factor is set to:

    * 1 second the successive sleeps will be 0.5, 1, 2, 4, 8, 16, 32, 64, 128,
        256.
    * 2 seconds - 1, 2, 4, 8, 16, 32, 64, 128, 256, 512
    * 10 seconds - 5, 10, 20, 40, 80, 160, 320, 640, 1280, 2560

    :see also:
        https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html

    """

    # noqa: RST304 Unknown interpreted text role "class".
    default_max_retries: ClassVar[int] = 10 # pylint: disable=E1136
    default_backoff: ClassVar[int] = 1 # pylint: disable=E1136
    max_retries: int = field(default_factory=generate_default_max_replies)
    backoff: int = field(default_factory=generate_default_backoff)
    session: Session = field(init=False)

    def __post_init__(self) -> NoReturn:
        """Postinitialize the Http class.

        During, the postinitialization, the retry strategy is configured and
        create a session with this retry strategy.
        """
        log.debug(
            'will use %s retries and a backoff factore of %s seconds',
            self.max_retries,
            self.backoff,
        )
        retry_strategy = Retry(
            total=self.max_retries,
            backoff_factor=self.backoff,
            status_forcelist=[401, 413, 429, 500, 502, 503, 504],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        # this is post init in order to create the session
        self.session = Session()  # noqa: WPS601 Found shadowed class attribute
        self.session.mount('https://', adapter)
        self.session.mount('http://', adapter)
